/*###########################################################################################################################
# Copyright (c) 1997-2012 Ufasoft   http://ufasoft.com   mailto:support@ufasoft.com                                         #
#                                                                                                                           #
# This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License #
# as published by the Free Software Foundation; either version 3, or (at your option) any later version.                    #                                                          #
#                                                                                                                           #
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied        #
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.     #
#                                                                                                                           #
# You should have received a copy of the GNU General Public License along with this program;                                #
# If not, see <http://www.gnu.org/licenses/>                                                                                #
###########################################################################################################################*/

#pragma once


namespace Ext {

class CCriticalSection;
class Exc;

class CPrintable {
public:
	virtual ~CPrintable() {}
	virtual String ToString() const =0;
};

typedef const Exc& RCExc;

#ifndef WDM_DRIVER

class EXTAPI CTls {	
public:
	typedef CTls class_type;

	CTls();
	~CTls();

	__forceinline void *get_Value() const {
#if UCFG_USE_PTHREADS
		return ::pthread_getspecific(m_key);
#else
		return ::TlsGetValue(m_key);
#endif
	}

	void put_Value(const void *p);
	DEFPROP(void*, Value);

private:

#if UCFG_USE_PTHREADS
	pthread_key_t m_key;
#else
	unsigned long m_key; //!!! was ULONG
#endif
};

template <class T>
class single_tls_ptr : protected CTls {
	typedef CTls base;
public:
	single_tls_ptr(T *p = 0) {
		base::put_Value(p);
	}

	single_tls_ptr& operator=(const T* p) {
		base::put_Value(p);
		return _self;
	}

	operator T*() const { return (T*)get_Value(); }
	T* operator->() const { return operator T*(); }
};

#if UCFG_USE_DECLSPEC_THREAD
#	define EXT_THREAD_PTR(typ, name) __declspec(thread) typ *name
#else
#	define EXT_THREAD_PTR(typ, name) Ext::single_tls_ptr<typ> name
#endif


#endif // !WDM_DRIVER

//!!!R const wchar_t * const NULL_STR = 0;
//!!!#define NULL_STR nullptr


class CStackTrace : public CPrintable {
public:
	EXT_DATA static bool Use;

	int AddrSize;
	std::vector<UInt64> Frames;

	CStackTrace()
		:	AddrSize(sizeof(void*))
	{}

	static CStackTrace AFXAPI FromCurrentThread();
	String ToString() const;
};

class Exc : public CPrintable {
public:
	typedef Exc class_type;

	static tls_ptr<String> t_LastStringArg;

	HRESULT HResult;
	String m_message;

	typedef std::map<String, String> CDataMap;
	CDataMap Data;

	explicit Exc(HRESULT hr = 0, RCString message = "");
	String ToString() const;
	virtual Exc *Clone() const { return new Exc(*this); }	

	virtual String get_Message() const;
	DEFPROP_VIRTUAL_GET_CONST(String, Message);
private:

#ifdef _WIN64
	LONG dummy;
#endif

public:
	CStackTrace StackTrace;
};

class ExcLastStringArgKeeper {
public:
	String m_prev;

	ExcLastStringArgKeeper(RCString s)
#if !UCFG_WDM
		:	m_prev(*Exc::t_LastStringArg)
#endif
	{
#if !UCFG_WDM
		*Exc::t_LastStringArg = s;
#endif
	}

	~ExcLastStringArgKeeper() {
#if !UCFG_WDM
		*Exc::t_LastStringArg = m_prev;
#endif
	}

	operator const String&() const {
#if !UCFG_WDM
		return *Exc::t_LastStringArg;
#endif
	}

	operator const char *() const {
#if !UCFG_WDM
		return *Exc::t_LastStringArg;
#endif
	}

	operator const String::Char *() const {
#if !UCFG_WDM
		return *Exc::t_LastStringArg;
#endif
	}
};

class OutOfMemoryExc : public Exc {
public:
	OutOfMemoryExc()
		:	Exc(E_OUTOFMEMORY)
	{
	}
};

class ArithmeticExc : public Exc {
public:
	ArithmeticExc(HRESULT hr = E_EXT_Arithmetic)
		:	Exc(hr)
	{}
};

class OverflowExc : public ArithmeticExc {
public:
	OverflowExc(HRESULT hr = E_EXT_Overflow)
		:	ArithmeticExc(hr)
	{}
};


class ArgumentExc : public Exc {
public:
	ArgumentExc()
		:	Exc(E_INVALIDARG)
	{}
};

class NotImplementedExc : public Exc {
public:
	NotImplementedExc()
		:	Exc(E_NOTIMPL)
	{}
};

class StackOverflowExc : public Exc {
	typedef Exc base;
public:
	StackOverflowExc()
		:	Exc(HRESULT_FROM_WIN32(ERROR_STACK_OVERFLOW))
	{}

	CInt<void *> StackAddress;
};

class AccessDeniedExc : public Exc {
public:
	AccessDeniedExc()
		:	Exc(E_ACCESSDENIED)
	{
	}
};

class ProcNotFoundExc : public Exc {
	typedef Exc base;
public:
	String ProcName;

	ProcNotFoundExc()
		:	Exc(HRESULT_OF_WIN32(ERROR_PROC_NOT_FOUND))
	{
	}

	String get_Message() const {
		String r = base::get_Message();
		if (!ProcName.IsEmpty())
			r += " "+ProcName;
		return r;
	}
};

INT_PTR WINAPI AfxApiNotFound();
#define DEF_DELAYLOAD_THROW static FARPROC WINAPI DliFailedHook(unsigned dliNotify, PDelayLoadInfo  pdli) { return &AfxApiNotFound; } static struct InitDliFailureHook {  InitDliFailureHook() { __pfnDliFailureHook2 = &DliFailedHook; } } s_initDliFailedHook;

extern "C" AFX_API void _cdecl AfxTestEHsStub(void *prevFrame);

//!!! DBG_LOCAL_IGNORE_NAME(1, ignOne);	

#define DEF_TEST_EHS							\
	static DECLSPEC_NOINLINE int AfxTestEHs(void *p) {			\
		try {								\
			AfxTestEHsStub(&p);					\
		} catch (...) {							\
		}										\
		return 0;								\
	}											\
	static int s_testEHs  = AfxTestEHs(&AfxTestEHs);									\


class AssertFailedExc : public Exc {
public:
	String FileName;
	int LineNumber;

	AssertFailedExc(RCString exp, RCString fileName, int line)
		:	Exc(HRESULT_FROM_WIN32(ERROR_ASSERTION_FAILURE), "Assertion Failed: "+exp)
		,	FileName(fileName)
		,	LineNumber(line)
	{
	}
};




template <typename H>
struct HandleTraitsBase {
	typedef H handle_type;

	static bool ResourceValid(const handle_type& h) {
		return h != 0;
	}

	static handle_type ResourseNull() {
		return 0;
	}
};

template <typename H>
struct HandleTraits : HandleTraitsBase<H> {
};

/*!!!
template <typename H>
inline void ResourceRelease(const H& h) {
	Throw(E_NOTIMPL);
}*/

template <typename H>
class ResourceWrapper {
public:
	typedef HandleTraits<H> traits_type;
	typedef typename HandleTraits<H>::handle_type handle_type;

	handle_type m_h;		//!!! should be private


	ResourceWrapper()
		:	m_h(traits_type::ResourseNull())
	{}

	ResourceWrapper(const ResourceWrapper& res)
		:	m_h(traits_type::ResourseNull())
	{
		operator=(res);
	}

	~ResourceWrapper() {
		if (Valid()) {
			if (!std::uncaught_exception())
				traits_type::ResourceRelease(m_h);
			else {
				try {
					traits_type::ResourceRelease(m_h);
				} catch (RCExc& DBG_PARAM(e)) {
					TRC(0, e);
				}
			}
		}
	}

	bool Valid() const {
		return traits_type::ResourceValid(m_h);
	}

	void Close() {
		if (Valid())
			traits_type::ResourceRelease(SwapRet(m_h, traits_type::ResourseNull()));
	}

	operator H() const {
		if (!Valid())
			Throw(E_HANDLE);
		return m_h;
	}

	ResourceWrapper& operator=(const ResourceWrapper& res) {
		handle_type tmp = SwapRet(m_h, traits_type::ResourseNull());
		if (traits_type::ResourceValid(tmp))
			traits_type::ResourceRelease(tmp);
		if (traits_type::ResourceValid(res.m_h))
			m_h = traits_type::Retain(res.m_h);
		return *this;
	}

	handle_type& OutRef() {
		if (Valid())
			Throw(E_EXT_AlreadyOpened);
		return m_h;
	}
};


template <class T>
class CHandleBase {
public:
	mutable volatile Int32 m_nInUse;

	CHandleBase()
		:	m_bClosed(true)
		,	m_nInUse(0)
	{
	}

	bool Release() const {
		Interlocked::Decrement(m_nInUse);
		if (!Interlocked::CompareExchange(m_nInUse, 0x80000000L, 0L)) {
			static_cast<const T*>(this)->InternalReleaseHandle();
			return true;
		}
		return false;
	}

	bool Close(bool bFromDtor = false) {
		if (!Interlocked::CompareExchange(m_bClosed, 1L, 0L)) {
			bool r = false;
#if !defined(_MSC_VER) || defined(_CPPUNWIND)
			if (!bFromDtor || !std::uncaught_exception())
				r = Release();
			else {
				try {
					r = Release();
				} catch (RCExc& DBG_PARAM(e)) {
					TRC(0, e);
				}
			}
#else
			r = Release();
#endif
			return r;
		}
		return false;
	}

protected:
	volatile Int32 m_bClosed;					// Int32 because we need Interlocked operations, but sizeof(bool)==1

template <class U> friend class CHandleKeeper;
};

class EXTAPI SafeHandle : public Object, public CHandleBase<SafeHandle> {
public:
	typedef HANDLE handle_type;

#ifndef WDM_DRIVER
	//!!!R static CTls t_pCurrentHandle;
	static EXT_THREAD_PTR(void, t_pCurrentHandle);
#endif

	CBool m_bOwn;	

	SafeHandle()
		:	m_handle(INVALID_HANDLE_VALUE)
		,	m_invalidHandleValue(INVALID_HANDLE_VALUE)
//!!!R		,	m_bClosed(true)
//!!!R		,	m_nInUse(0)
		,	m_bOwn(true)
	{}

	SafeHandle(HANDLE invalidHandle, bool)
		:	m_handle(invalidHandle)
		,	m_invalidHandleValue(invalidHandle)
//!!!R		,	m_bClosed(true)
//!!!R		,	m_nInUse(0)
		,	m_bOwn(true)
#if UCFG_WDM
		,	m_pObject(nullptr)
#endif
	{}

	SafeHandle(HANDLE handle);
	virtual ~SafeHandle();
//!!!	void Release() const;
	//!!!  void CloseHandle();
	HANDLE DangerousGetHandle() const;
	void Attach(HANDLE handle);

	EXPLICIT_OPERATOR_BOOL() const {
		return Valid() ? EXT_CONVERTIBLE_TO_TRUE : 0;
	}

#if UCFG_WDM
protected:
	mutable void *m_pObject;
	mutable CBool m_CreatedReference;
	void AttachKernelObject(void *pObj, bool bKeepRef);
public:
	void Attach(PKEVENT pObj) { AttachKernelObject(pObj, false); }
	void Attach(PKMUTEX pObj) { AttachKernelObject(pObj, false); }

	NTSTATUS InitFromHandle(HANDLE h, ACCESS_MASK DesiredAccess, POBJECT_TYPE ObjectType, KPROCESSOR_MODE  AccessMode);
#endif


	HANDLE Detach();
	void Duplicate(HANDLE h, DWORD dwOptions = 0);
	virtual bool Valid() const;

	class HandleAccess : public CHandleKeeper<SafeHandle> {
		typedef CHandleKeeper<SafeHandle> base;
	public:
		HandleAccess()
		{}

		HandleAccess(const HandleAccess& ha)
			:	base(ha)
		{}

		HandleAccess(const SafeHandle& h)
			:	base(h)
		{}

		~HandleAccess() {	
#if !defined(_MSC_VER) || defined(_CPPUNWIND)
			try //!!!
#endif
			{
				Release();
			}
#if !defined(_MSC_VER) || defined(_CPPUNWIND)
			catch (RCExc) {
			}
#endif
		}

		operator SafeHandle::handle_type() {
			if (!m_bIncremented)
				return 0;
			return m_hp->DangerousGetHandle();
		}
	};

	class BlockingHandleAccess : public HandleAccess {
	public:
		BlockingHandleAccess(const SafeHandle& h);
		~BlockingHandleAccess();
	private:
		HandleAccess *m_pPrev;
	};

	void InternalReleaseHandle() const;

	/*!!!R	__forceinline operator HANDLE() const
	{
	return m_handle;
	} */        
protected:

	HANDLE DangerousGetHandleEx() const { return m_handle; }
	virtual void ReleaseHandle(HANDLE h) const;	

private:
	mutable HANDLE m_handle;
//!!!R	volatile Int32 m_bClosed; // Int32 because we need Interlocked operations, but sizeof(bool)==1
	const HANDLE m_invalidHandleValue;
//!!!R	mutable volatile Int32 m_nInUse;


	SafeHandle& operator=(const SafeHandle& h) { //!!! Hide
		THROW_UNK();
	}

	template <class T> friend class CHandleKeeper;
};

template <typename T>
typename T::handle_type Handle(T& x) {
	return T::HandleAccess(x);
}

template <typename T>
typename T::handle_type BlockingHandle(T& x) {
	return T::BlockingHandleAccess(x);
}

class CEscape {
public:
	virtual ~CEscape() {}

	virtual void EscapeChar(std::ostream& os, char ch) {
		os.put(ch);
	}

	virtual int UnescapeChar(std::istream& is) {
		return is.get();
	}

	static AFX_API String AFXAPI Escape(CEscape& esc, RCString s);
	static AFX_API String AFXAPI Unescape(CEscape& esc, RCString s);
};

struct CExceptionFabric {
	CExceptionFabric(int facility);
	virtual DECLSPEC_NORETURN void ThrowException(HRESULT hr, RCString msg) =0;
};

typedef std::map<int, CExceptionFabric*> CExceptionFabrics; 
extern EXT_DATA CExceptionFabrics g_exceptionFabrics;

DECLSPEC_NORETURN void AFXAPI ThrowS(HRESULT hr, RCString msg);



} // Ext::

