/*###########################################################################################################################
# Copyright (c) 1997-2012 Ufasoft   http://ufasoft.com   mailto:support@ufasoft.com                                         #
#                                                                                                                           #
# This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License #
# as published by the Free Software Foundation; either version 3, or (at your option) any later version.                    #                                                          #
#                                                                                                                           #
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied        #
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.     #
#                                                                                                                           #
# You should have received a copy of the GNU General Public License along with this program;                                #
# If not, see <http://www.gnu.org/licenses/>                                                                                #
###########################################################################################################################*/

// Support of device "BitForce SHA256 Single"  http://butterflylabs.com/products/

#include <el/ext.h>

#if UCFG_USE_POSIX
#	include <termios.h>

class CTermiosKeeper {
public:
	int m_fd;
	termios m_orig;

	CTermiosKeeper(int fd)
		:	m_fd(fd)
	{
		CCheck(::tcgetattr(m_fd, &m_orig));
	}

	~CTermiosKeeper() {
		::tcsetattr(m_fd, 0, &m_orig);
	}
};

#endif


#include "miner.h"

namespace Coin {

class BitforceThread : public WorkerThreadBase {
	typedef WorkerThreadBase base;
public:
	String m_devpath;
	CBool Detected;
	AutoResetEvent m_evDetected, m_evStartMining;

	BitforceThread(BitcoinMiner& miner, CThreadRef *tr)
		:	base(miner, tr)
	{
		m_bSleeping = true;
	}
protected:
	String GetReply(FILE *file) {
		char buf[100];
		if (!fgets(buf, sizeof buf, file))
			CCheck(-1);
		fseek(file, 0, SEEK_CUR);					// dont check retval, switch after reading
		String r = String(buf).Trim();
		TRC(3, "Reply: " << r);
		return r;
	}

	String Command(FILE *file, const char *cmd, size_t size = 3) {
		if (size == 3) {
			TRC(3, cmd);
		}
		CCheck(fwrite(cmd, 1, size, file));
		CCheck(fflush(file));
		fseek(file, 0, SEEK_CUR);				// dont check retval, switch after reading, bug in the MS CRT fflush() for non-buffered streams;
		return GetReply(file);
	}

	void Execute() override {
		Name = "BitforceThread";

		if (FILE *file = fopen(m_devpath, "r+b")) {
			try {			
				struct FileKeeper {
					FILE *m_file;

					~FileKeeper() {
						fclose(m_file);
					}
				} fileKeeper = { file };			
				CCheck(setvbuf(file, 0, _IONBF, 0));
	#if UCFG_USE_POSIX
				CTermiosKeeper termKeeper(fileno(file));
				termios tio = termKeeper.m_orig;
				::cfmakeraw(&tio);
				CCheck(::tcsetattr(termKeeper.m_fd, 0, &tio));
	#endif
				UInt32 portion = FULL_TASK;

				String id = Command(file, "ZGX");
				if (id == ">>>ID: BitFORCE SHA256 Version 1.0>>>") {
					Detected = true;
					m_evDetected.Set();
					m_evStartMining.Lock();
					char bufData[8+32+12+8];
					memset(bufData, '>', sizeof bufData);
					for (BitcoinWorkData wd; !m_bStop && (wd = Miner.GetWorkForThread(_self, portion));) {
						StartRound();
						if (Command(file, "ZDX") != "OK") {
							TRACE_BITCOIN(cerr << Name << " signaled Error" << endl);
							break;
						}
						memcpy(bufData+8, wd.Midstate.constData(), 32);
						memcpy(bufData+8+32, wd.Data.constData()+64, 12);
						TRC(3, "Data[64..75]: " << ConstBuf(wd.Data.constData()+64, 12) << ", Midstate: " << wd.Midstate);
						if (Command(file, bufData, sizeof bufData) != "OK") {
							TRACE_BITCOIN(cerr << Name << " signaled Error" << endl);
							break;						
						}
						String reply;
						while (!m_bStop && (reply = Command(file, "ZFX")) == "BUSY")
							Thread::Sleep(10);
						if (m_bStop)
							break;
						TRC(3, "Reply: " << reply);
						if (reply.Left(12) == "NONCE-FOUND:") {
							vector<String> nn = reply.Substring(12).Split(",");
							for (int i=0; i<nn.size(); ++i)
								Miner.TestAndSubmit(_self, wd, Convert::ToUInt32(nn[i], 16));
						} else if (reply != "NO-NONCE") {
							TRACE_BITCOIN(cerr << Name << " Unexpected reply: " << reply << endl);
							break;
						}
						Interlocked::Add(Miner.HashCount, portion);
						CompleteRound(0x100000000ULL);
					}
				}
			} catch (RCExc e) {
				TRACE_BITCOIN(cerr << e.Message << endl);
			} catch (...) {
				TRACE_BITCOIN(cerr << "Unknown exception detected" << endl);
			}			
		}
		m_evDetected.Set();
	}

	void Stop() override {
		base::Stop();
		m_evStartMining.Set();
	}
};


class BitforceDevice : public ComputationDevice {
public:
	ptr<BitforceThread> m_thread;

	void Start(BitcoinMiner& miner, CThreadRef *tr) override {
		m_thread->m_owner = tr;
		miner.Threads.push_back(m_thread.get());
		m_thread->m_evStartMining.Set();
	}
};

class BitforceArchitecture : public ComputationArchitecture {
public:

	void FillDevices(BitcoinMiner& miner) override {
#if UCFG_USE_POSIX
		vector<String> devpaths = Directory::GetFiles("/dev", "ttyUSB*");
		for (int i=0; i<devpaths.size(); ++i) {
			ptr<BitforceThread> t = new BitforceThread(miner, 0);
			t->m_devpath = devpaths[i];
			t->Start();
			if (t->m_evDetected.Lock(1000)) {
				if (t->Detected) {
					ptr<BitforceDevice> dev = new BitforceDevice;
					dev->Description = "BitForce SHA256 Single";
					dev->Name = t->m_devpath + " ["+dev->Description+"]";
					dev->m_thread = t;
					miner.Devices.push_back(dev.get());
				}
			}
		}
#endif	// UCFG_USE_POSIX
	}
} g_bitforceArchitecture;

} // Coin::






