/*###########################################################################################################################
# Copyright (c) 1997-2012 Ufasoft   http://ufasoft.com   mailto:support@ufasoft.com                                         #
#                                                                                                                           #
# This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License #
# as published by the Free Software Foundation; either version 3, or (at your option) any later version.                    #                                                          #
#                                                                                                                           #
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied        #
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.     #
#                                                                                                                           #
# You should have received a copy of the GNU General Public License along with this program;                                #
# If not, see <http://www.gnu.org/licenses/>                                                                                #
###########################################################################################################################*/

#pragma once

#include <el/libext/ext-http.h>

#include "bitcoin-common.h"
#include "bitcoin-sha256.h"

namespace Ext {
	namespace Cl {
	class Device;
}}

namespace Coin {

class BitcoinMiner;

const float BITCOIN_MAX_GPU_TEMPERATURE = 83;
const int BITCOIN_DEFAULT_GPU_IDLE_MILLISECONDS = 100;

const UInt32 FULL_TASK = UInt32(0x100000000ULL / UCFG_BITCOIN_NPAR);

const int MIN_GETWORK_QUEUE = 3,
			MAX_GETWORK_QUEUE = 5;

const int NORMAL_WAIT = 2000;


class BitcoinWorkData {
public:
	DateTime Timestamp;
	Blob Midstate, Data, Target, Hash1;
	UInt32 FirstNonce, LastNonce;
	UInt16 RollNTime;

	BitcoinWorkData()
		:	FirstNonce(0)
		,	LastNonce(UInt32(-1))
		,	RollNTime(0)
	{}

	EXPLICIT_OPERATOR_BOOL() const {
		return Data.Size != 0 ? EXT_CONVERTIBLE_TO_TRUE : 0;
	}

	bool IsFull() const {
		return FirstNonce==0 && LastNonce == UInt32(-1);
	}

	static BitcoinWorkData FromJson(RCString json);
	static BitcoinWorkData GetTestData();
	void SetRollNTime(WebClient& wc);
};

class ComputationDevice : public Object {
public:
	String Name, Description;
	CBool IsAmdGpu;
//!!!	int Index;
	ptr<Object> Miner;

	virtual void Start(BitcoinMiner& miner, CThreadRef *tr) =0;
};

class ComputationArchitecture {
public:
	static ComputationArchitecture *s_pChain;
	ComputationArchitecture *m_pNext;

	ComputationArchitecture()
		:	m_pNext(SwapRet(s_pChain, this))
	{
	}

	virtual void FillDevices(BitcoinMiner& miner) =0;
};

class WorkerThreadBase : public Thread {
	typedef Thread base;
public:
	BitcoinMiner& Miner;
	Int64 m_prevTsc;
	UInt32 DevicePortion;
	ptr<ComputationDevice> Device;

	WorkerThreadBase(BitcoinMiner& miner, CThreadRef *tr)
		:	base(tr)
		,	Miner(miner)
		,	CurrentWebClient(0)
		,	m_nHashes(0)
		,	m_prevTsc(0)
		,	DevicePortion(128)
	{
		m_bSleeping = true;
	}

	virtual String GetDeviceName() { return nullptr; };
protected:
	WebClient *CurrentWebClient;
	TimeSpan m_span;
	UInt64 m_nHashes;

#if UCFG_WIN32
	void OnAPC() override {
		base::OnAPC();
		if (CurrentWebClient)
			CurrentWebClient->CurrentRequest->ReleaseFromAPC();
	}
#endif

	void Stop() override;
	void CompleteRound(UInt64 nHashes);

	void StartRound() {
		m_dtStartRound = DateTime::UtcNow();
	}
private:
	DateTime m_dtStartRound;

	friend class BitcoinMiner;
};

class CpuDevice : public ComputationDevice {
public:
	void Start(BitcoinMiner& miner, CThreadRef *tr) override;
};

class GpuDevice : public ComputationDevice {
};

class AmdGpuDevice : public GpuDevice {
public:
	AmdGpuDevice() {
		IsAmdGpu = true;
	}

	void Start(BitcoinMiner& miner, CThreadRef *tr) override;
};

#if UCFG_BITCOIN_USE_CAL
class GpuThreadBase : public WorkerThreadBase {
	typedef WorkerThreadBase base;
public:
	GpuThreadBase(BitcoinMiner& miner, CThreadRef *tr)
		:	base(miner, tr)
	{}

};

#endif

struct SUrlTtr {
	String Url;
	int TtrMinutes;

	bool operator==(const SUrlTtr& x) const {
		return Url==x.Url && TtrMinutes==x.TtrMinutes;
	}

	bool operator!=(const SUrlTtr& x) const {
		return !operator==(x);
	}
};

class BitcoinWebClient : public WebClient {
	typedef WebClient base;

public:
	int Timeout;

	BitcoinWebClient()
		:	Timeout(-1)
	{}
protected:
	void OnHttpWebRequest(HttpWebRequest& req) {
		base::OnHttpWebRequest(req);
		if (Timeout >= 0)
			req.Timeout = Timeout;
	}
};

class MINER_CLASS BitcoinMiner {
public:
//	static const Int32 NONCE_STEP = 128*1024*1024; //!!! was 8
	
	int GetworkPeriod;
	float Speed;		// float to be atomary
	Int64 EntireHashCount;
	DateTime DtStart;

	mutex m_csGetWork;
	ManualResetEvent m_evDataReady;
	AutoResetEvent m_evGetWork;

	int GpuIdleMilliseconds;
	CBool m_bTryGpu, m_bLongPolling;
	vector<ptr<ComputationDevice>> Devices;

	int m_msWait;
	int NPAR;
	volatile Int32 HashCount;
	volatile Int32 AcceptedCount;
	int Verbosity;
	String UserAgentString;

	DateTime m_dtNextGetWork;

	String MainBitcoinUrl, Login, Password;
	DateTime m_dtSwichToMainUrl;
	String ProxyString;

	int ThreadCount;
	CPointer<CThreadRef> m_tr;
	vector<ptr<Thread> > Threads;
	ptr<Thread> GetWorkThread;
	ptr<Thread> GpuThread;
	int m_minGetworkQueue; 
	float MaxTemperature;

	CBool m_bStop;
	CBool m_bInited;

	BitcoinMiner();
	void InitDevices();
	BitcoinWorkData GetWorkForThread(WorkerThreadBase& wt, UInt32 portion);
	virtual BitcoinWebClient GetWebClient(WorkerThreadBase *wt);
//!!!	void SetNewData(const BitcoinWorkData& wd);
	void SetNewData(const BitcoinWorkData& wd, bool bClearOld = false);
	void SetWebInfo(WebClient& wc);
	void CheckLongPolling(WebClient& wc);
	void CallNotifyRequest(WorkerThreadBase& wt, RCString url);	
	
	virtual BitcoinWorkData GetWork(WebClient*& curWebClient);
	virtual bool SubmitResult(WebClient*& curWebClient, const BitcoinWorkData& wd);

	bool TestAndSubmit(WorkerThreadBase& wt, BitcoinWorkData& wd, UInt32 nonce);
	virtual String GetCalIlCode(bool bEvergreen);
	virtual String GetOpenclCode();
	virtual String GetCudaCode();

	bool UseSse2() {
#if UCFG_CPU_X86_X64
		return CpuInfo().HasSse2;
#endif
		return false;
	}

	void Start(CThreadRef *tr = nullptr);
	void Stop();
	virtual void OnRoundComplete() {}
private:
	mutex m_csCurrentData;
	
	typedef list<BitcoinWorkData> CTaskQueue;
	CTaskQueue TaskQueue;

	list<BitcoinWorkData> WorksToSubmit;

	String BitcoinUrl;
	vector<SUrlTtr> HostList;
	CInt<int> HostIndex;

	String GetCurrentUrl();

	mutex m_csLongPollingThread;
	ptr<Thread> m_LongPollingThread;

	friend class WorkerThread;
	friend class LongPollingThread;
	friend class GetWorkThread;
};

class GpuMiner;

class GpuTask : public BitcoinSha256 {
public:
	BitcoinWorkData WorkData;
	WorkerThreadBase *m_pwt;
	int m_npar, m_nparOrig;
	GpuMiner& m_gpuMiner;

	GpuTask(GpuMiner& gpuMiner)
		:	m_gpuMiner(gpuMiner)
	{}

	virtual void Prepare(const BitcoinWorkData& wd, WorkerThreadBase *pwt) =0;
	virtual void Run() =0;
	virtual bool IsComplete() =0;
	virtual void Wait() =0;
	virtual int ParseResults() =0;
};

class GpuMiner : public Object {
public:
	BitcoinMiner& Miner;
	UInt32 DevicePortion;

	GpuMiner(BitcoinMiner& miner)
		:	Miner(miner)
		,	DevicePortion(1024*1024/UCFG_BITCOIN_NPAR)
	{}

//!!!	virtual void InitBeforeRun() {}
	virtual ptr<GpuTask> CreateTask() =0;
	virtual int Run(const BitcoinWorkData& wd, WorkerThreadBase *pwt) =0;
};

ptr<GpuMiner> CreateOpenclMiner(BitcoinMiner& miner, Cl::Device& dev, const ConstBuf& binary);

} // Coin::

